#ifndef __STDC_LIMIT_MACROS
#define __STDC_LIMIT_MACROS
#endif

#include "Pedigree.h"
#include "Parameters.h"
#include "GenomeSequence.h"
#include "Random.h"
#include <ctype.h>
#include "Parameters.h"
#include "IntHash.h"
#include "MemoryMapArray.h"
#include <map>
#include "PedigreeAlleles.h"
#include <stdio.h>
#include <stdlib.h>
#include "readbam.h"
#include "MemoryAllocators.h"
#include "math.h"

using namespace std;



#define DELIM1 " \t"
#define DELIM "\t"
#define SAM_READ_COLUMN 9
#define SAM_QUALITY_COLUMN 10
#define SAM_CHROMOSOME_COLUMN 2
#define SAM_CHROMOSOME_INDEX_COLUMN 3
#define SAM_MAPQ_COLUMN 4
#define SAM_CIGAR_COLUMN 5


#define HAPMAP_MATCH_COLUMN 1
#define HAPMAP_MISMATCH_COLUMN 2
#define NONDBSNP_MATCH_COLUMN 4
#define NONDBSNP_MISMATCH_COLUMN 5

#define READLOWERBOUND 5000

void error(int flagSAM)
{ if (flagSAM) printf("It does not look like a right SAM format file!");
  else printf("It does not look like a right BAM format file!");
}

void usage(int argc, char **argv)
{
#ifndef VERSION
	printf("LaneCheck 0.0.1 (c) 2009.7 Wei Chen, Paul Anderson, Goncalo Abecasis\n");
#else 
	printf("LaneCheck " VERSION " (c) 2009.7 Wei Chen, Paul Anderson, Goncalo Abecasis\n");
#endif
}


int MatchAllele(char readbase, char alleleNumber)
{   char allele = alleleNumber;
    switch (alleleNumber)
      { case '1': allele='a';break;
        case '2': allele='c';break;
        case '3': allele='g';break;
        case '4': allele='t';break;
        default: allele = tolower(alleleNumber); break;
      }
    if (readbase == allele) return 1;
    else  if ((readbase == 'a' || readbase == 'c' || readbase == 'g' || readbase == 't')&& 
              (allele == 'a' || allele == 'c' || allele == 'g' || allele == 't')) return 0;
          else return -1;
      
}
 

int PrintSummary(FILE *fout, double **summaryTable, double *summaryCounts, String lanefile, String fid, String pid)
{ 		
		
		for (int i = 0; i < 100; i ++)
		{ 
			int tmpCount1 = summaryTable[i][1] + summaryTable[i][2];
			int tmpCount2 = summaryTable[i][4] + summaryTable[i][5];

			summaryCounts[1] += summaryTable[i][1];
			summaryCounts[2] += summaryTable[i][2];
			summaryCounts[4] += summaryTable[i][4];
			summaryCounts[5] += summaryTable[i][5];

			if (tmpCount2 > 0)
				summaryTable[i][6] = summaryTable[i][5]/tmpCount2;

			if (tmpCount1 > 0)
				summaryTable[i][3] = summaryTable[i][2]/tmpCount1; summaryCounts[6] += tmpCount1 * summaryTable[i][6];

		}

		summaryCounts[3] = summaryCounts[2]/(summaryCounts[1]+summaryCounts[2]);
		summaryCounts[6] /= (summaryCounts[1] + summaryCounts[2]);
		summaryCounts[7] =  summaryCounts[3] - summaryCounts[6];
		
		// Output summary statistics //  
					 
		for (int i = 0; i < 100; i ++) {
			for (int j = 0; j < 7; j ++) {
				if (j == 3 || j == 6)
					fprintf(fout,"%7.6lf\t", summaryTable[i][j]);
				else
					fprintf(fout,"%d\t",(int)summaryTable[i][j]);
			
		  }
		  fprintf(fout,"\n");
		}
		 
		fprintf(fout,"CHECKRESULT\t%s\t%s\t%s\t", (const char*) lanefile, (const char*) fid, (const char*) pid);
		
		for (int i = 1; i < 8; i ++) {
			if (i == 3 || i == 6 || i == 7)
				fprintf(fout,"%7.6lf\t", summaryCounts[i]);
			else
				fprintf(fout,"%d\t", (int)summaryCounts[i]);
		}
		fprintf(fout,"\n");
        
        return 1;
    
    
}       
        

int main (int argc, char ** argv)
{  
	String readFile;
    String keyFile;
	String reference; 
	String dbSnpFile;
	//String sampleFile;
	String pedFile;
	String datFile;
	String mapFile;
	String outFile;
    int mapq = -1;
    int maxcount = 1000000;
    bool nonrefhomo = false;
    bool verbose = false;
    bool coverage = false;
    bool countbysite = false;
    bool memoryMap = false;
    
    
    dbSnpFile = "/home/1000G/data/GenomeSNP.dbsnp";
    reference= "/usr/cluster/share/karma/NCBI36.fa";


	ParameterList pl;

	BEGIN_LONG_PARAMETERS(parameters)
		LONG_PARAMETER_GROUP("Input Files")
			LONG_STRINGPARAMETER("referencegenome", &reference)
			LONG_STRINGPARAMETER("dbSNPfile", &dbSnpFile)
			LONG_STRINGPARAMETER("LaneSamplefile", &keyFile)
			LONG_STRINGPARAMETER("pedfile", &pedFile)
			LONG_STRINGPARAMETER("datfile", &datFile)
			LONG_STRINGPARAMETER("mapfile", &mapFile)
            LONG_INTPARAMETER("mapquality", &mapq)
            LONG_INTPARAMETER("genocount", &maxcount)
            LONG_PARAMETER("nonrefhomo", &nonrefhomo)  // 0  nonreference homo sites, 1 reference homo sites  
            LONG_PARAMETER("verbose",&verbose)
            LONG_PARAMETER("coverage",&coverage)
            LONG_PARAMETER("countbysite",&countbysite)
            LONG_PARAMETER("memorymap",&memoryMap)
            
		LONG_PARAMETER_GROUP("Output Files")
			LONG_STRINGPARAMETER("prefix", &outFile)
	
	END_LONG_PARAMETERS();
	
	pl.Add(new HiddenString('d', "datfile", datFile));
	pl.Add(new HiddenString('p', "pedfile", pedFile));
	pl.Add(new HiddenString('i', "mapfile", mapFile));
    
    
	
	pl.Add(new LongParameters("Availabe Options",parameters));
	
	usage(argc, argv);
	pl.Read(argc, argv);
	pl.Status(); 
    
    	
	printf("Loading pedigree:\n");
	
	Pedigree ped;
	printf("\tPrepare: ");
	ped.Prepare(datFile);
	printf("Done\n");
	printf("\tLoad: ");fflush(stdout);
	ped.Load(pedFile);
	printf("Done\n");
	printf("\tLoadMarkerMap: ");fflush(stdout);
	ped.LoadMarkerMap(mapFile);
	
	printf("Loading pedigree: DONE\n");
    
    
    printf("markerCount=%d\t",ped.markerCount);

	//printf("%d %d %d\n", ped.count,ped.markerCount,ped.traitCount);
	
	
	GenomeSequence referencegenome;
	 
	fprintf(stderr,"Open reference: ");
        
    referencegenome.setFastaName(reference);
    referencegenome.useMemoryMap(memoryMap);
    if(referencegenome.open()) {
		fprintf(stderr,"failed to open reference file.\n");
		exit(1);
	}
	fprintf(stderr, "DONE!\n");        
            
            
	
  // mask indicator of dbSNp and HapMap site  0 - normal, 1 - dbSNP  2 - HapMap 

	map<genomeIndex_t, int> snpPos2Index;
    
    
	for (int i = 0; i < ped.markerCount ; i ++) {
		int snpPos = round(Pedigree::GetMarkerInfo(i)->position * 100); 

        
        if (Pedigree::GetMarkerInfo(i)->chromosome >=1 && Pedigree::GetMarkerInfo(i)->chromosome <= 22){ 
		genomeIndex_t snpChrPos = referencegenome.getGenomePosition(
		Pedigree::GetMarkerInfo(i)->chromosome - 1, snpPos); // use int is much more efficient  only on autosomal 

		snpPos2Index.insert(pair<genomeIndex_t,int>(snpChrPos,i));
        }
	}

	//printf("index = %d\n",snpPos2Index[1677919547]);
    	

#define LANECHECK_MASKFILE "lanecheck.genomemask"	

	mmapArray2Bit_t mask;
    
    /*
	struct stat sb;
	if(stat(LANECHECK_MASKFILE, &sb) == 0) {
		printf("Open existing lane check mask file '%s': ",
				LANECHECK_MASKFILE);
		assert(mask.open(LANECHECK_MASKFILE)==0);
		printf("Done!\n");
	} else         
        */
        {

		printf("creating genome mask\n");

		// create the mask file...
		mmapArrayBool_t dbSnpVector;

        
		// we need the dbSnpFile data, too:
        dbSnpVector.useMemoryMap(memoryMap);  
        
		if(dbSnpVector.open(dbSnpFile)) {
			fprintf(stderr,"failed to open dbSNP file %s\n", dbSnpFile.c_str());
			exit(1);
		}
		 
        
                
        mask.useMemoryMap(false);       
		assert(mask.create(LANECHECK_MASKFILE, referencegenome.sequenceLength())==0);
		
        long int i;
		for ( i = 0; i < (long int)referencegenome.sequenceLength(); i ++) {
			mask.set(i, dbSnpVector[i]);
		}
        
        //printf("mask = %ld\n",i);
		//printf("mask = %ld %d \n",i, mask[i-1]);
        
		//
		for (int i = 0; i < ped.markerCount ; i ++) {
			int snpPos = round(Pedigree::GetMarkerInfo(i)->position * 100);

            
            if (Pedigree::GetMarkerInfo(i)->chromosome <=22 && Pedigree::GetMarkerInfo(i)->chromosome >=1) {  //ignore x  Pedigree::GetMarkerInfo(i)->chromosome gives 999 for chr X
            genomeIndex_t snpChrPos = referencegenome.getGenomePosition(
			 Pedigree::GetMarkerInfo(i)->chromosome - 1, snpPos); // use int is much more efficient

			 mask.set(snpChrPos, 2); 
            } 
		}

		dbSnpVector.close();
		
		printf("Creating genome mask: DONE\n");
	}
    
    
    
       

    StringArray tokens;
	String buffer;
          		
				
	IFILE f; 
	IFILE fkey = ifopen(keyFile,"r");
    if (fkey== NULL) {printf ("Can't open LaneSample file\n");exit(1);}
    
	
    if (outFile.IsEmpty()) outFile = "test";	
	FILE *fout = fopen(outFile+".out","w");
    
    if (fout == NULL) {printf ("Can't open output file\n");exit(1);}
	 
   	int sampleLine = 0;
	double **summaryTable = AllocateDoubleMatrix(300,7);
    double * summaryCounts  = new double [8];
    int *covarray = new int [ped.markerCount];
    int **maparray = AllocateIntMatrix (ped.markerCount,2);
    
    SAMFile bamfile;
    SAMEntry entry;
    String read;
    String quality;
    String chromosomeName;
    String cigar;
    int chromosomeIndex;
    //int mapquality;  
    String mapquality;
    int flagSAM;
    int count;   
    int endFile;
    int offset;
    int baseQuality;  
    String lanefile;
    String fid;
    String pid;
    String readID;
    genomeIndex_t readPosition;    
    CigarCol cigarcol;
    String aligTypes;        
    Person *currentSample;
    Alleles allele;
    MarkerInfo *info;
    char alleleNumber;
    char readchar;
    int matchallele;
    long int refpos;
    int genocount;
    
    
    
    //printf("%c\n",referencegenome[1677919547]);

            
    
    while (!ifeof(fkey)) {
    	sampleLine ++; printf ("Finish row = %d\n",sampleLine); fflush(stdout);

                       
		tokens.Clear();
		buffer.Clear();
		buffer.ReadLine(fkey);

		if (buffer.IsEmpty()) continue;
		tokens.AddTokens(buffer, DELIM1);

		if (tokens.Length() < 3) {
			fprintf(fout, "Line %d has incorrect format!\n",sampleLine);
			continue;
		}

        lanefile = tokens[0];
		fid = tokens[1];
		pid = tokens[2];   //printf("%s %s\n", (const char*) fid, (const char*) pid); 
		
        flagSAM = 0;
        
        
        if (lanefile.AsLower().Find("sam")>0) flagSAM = 1;
        else if (lanefile.AsLower().Find("bam")>0) flagSAM = 0;
             else printf("The file extension can not be recognized. It must be .sam or .bam!\n");
        
        
          
		currentSample = ped.FindPerson(fid, pid); 
		if (currentSample == NULL) {
			fprintf(fout,"%d: Can't find sample %s %s in the pedigree file!\n",
				sampleLine, (const char*) fid, (const char*) pid);
			continue;
		}

		count = 0;   genocount = 0;

		for (int i = 0; i < 100; i ++)
			for (int j = 0; j < 7; j ++) {
				if (j == 0) summaryTable[i][j] = i;
				else summaryTable[i][j] = 0;
			}
            
            
                        
          for (int i = 0; i < ped.markerCount; i ++) {   covarray[i] = 0; maparray[i][0] = 0; maparray[i][1] = 0;}
                 
          
          
          endFile = 0;   
                  
          
          if (flagSAM) {    f = ifopen(lanefile, "r");  
                           if (f == NULL) {printf ("Can't open lane file\n");exit(1);}
          }
          else   bamfile.Open(lanefile);   
                             
                   
          
          while (!endFile)  {  
                          
          if (flagSAM) {   
                             
              if (ifeof(f)) { endFile = 1; break;}
              
      	      tokens.Clear(); buffer.Clear();
			  buffer.ReadLine(f);
			  if (buffer.IsEmpty() || buffer[0] == '@') continue;
			  count ++;                                            
			  tokens.AddTokens(buffer, DELIM); 
              
              readID = tokens[0];
              read = tokens[SAM_READ_COLUMN];
			  quality = tokens[SAM_QUALITY_COLUMN];
			  chromosomeName = tokens[SAM_CHROMOSOME_COLUMN];
			  chromosomeIndex = tokens[SAM_CHROMOSOME_INDEX_COLUMN].AsInteger();
              cigar = tokens[SAM_CIGAR_COLUMN];
              mapquality = tokens[SAM_MAPQ_COLUMN];
              
              if (count == 1 && (!read.AsUpper().Find('A'))  && (!read.AsUpper().Find('C')) &&
                     (!read.AsUpper().Find('G')) && (!read.AsUpper().Find('T'))) error(flagSAM);     
            // if (count != 9) continue; 
            // printf("%s\t%s\t%s\t%s\n",(const char*)readID, (const char*)read,(const char*) cigar,(const char*) mapquality);
            
           }
           
           else {              
            count ++;   
            entry.Clear();  
            if (bamfile.ReadLine(entry)<0) { endFile = 1; break; }
                    
            readID = entry.QNAME;                                        
            read = entry.SEQ;
			quality = entry.QUAL;
			chromosomeName = entry.RNAME;
			cigar = entry.CIGAR;
            chromosomeIndex = entry.POS;            
            mapquality = entry.MAPQ;
                        
            if (count == 1 && (!read.AsUpper().Find('A'))  && (!read.AsUpper().Find('C')) &&
                     (!read.AsUpper().Find('G')) && (!read.AsUpper().Find('T'))) error(flagSAM);   
            
           
            
            }
            
            
            //printf("%s %s %s %d\n", (const char*)read, (const char*) quality, (const char*)chromosomeName,chromosomeIndex);
            
       		
			if(isdigit(chromosomeName[0])) {
				// The indexed version of getGenomePosition is far faster than the
				// character name lookup.
				//readPosition = referencegenome.getGenomePosition(chromosomeName.AsInteger() - 1, chromosomeIndex.AsInteger());
                readPosition = referencegenome.getGenomePosition(chromosomeName.AsInteger() - 1, chromosomeIndex);
			} else { continue;
				// Here we have for example 'X' or 'Y' chromosome -- just use the slow search for them.
				// It may be that we can just discard these cases - e.g. do a continue here.
				//readPosition = referencegenome.getGenomePosition(chromosomeName.c_str(), chromosomeIndex.AsInteger());
			}
        
            
            
            //if (mapquality < mapq) continue;
            //if (cigar.FindChar('I')>0 || cigar.FindChar('D') >0 || cigar.FindChar('S') > 0) continue;
             // FindChar  if found return postion, else return -1
            
                       
           cigarcol.CigarParse(cigar); 
                       
           if (cigarcol.fieldTypes[0] == 'S') { cigarcol.CigarClear(); continue;}
           
                       
            aligTypes="";
            for (int i = 0; i < cigarcol.size; i++)
               for (int j = 0; j < cigarcol.fieldCount[i]; j++)
            {  aligTypes += cigarcol.fieldTypes[i];
            }
            
            //printf("size = %d\t%d\t%d\n",aligTypes.Length(),read.Length(),quality.Length());
            
            cigarcol.CigarClear();  
                       
            offset = 0;
             
			for (int i = 0; i < read.Length(); i++) {
                                               				
                if (aligTypes[i] == 'S') continue;
                if (aligTypes[i] == 'I') { offset--; continue; }  
                if (aligTypes[i] == 'D') { offset++; continue; }
                
                              
                                               
                baseQuality = quality[i] - '!';
                
                
                
                
                if (baseQuality < 0 || baseQuality > 99) continue;
                
                
				// cigar MIDS[NHP]  e.g. 16M2I5D8M              
                
                refpos = readPosition + i + offset;     //printf("%d %ld\t",i,refpos);
 
	           // filter out dbSNP positions: need to consider more cases
				if(mask[refpos] == 0) {
                                    
					if(read[i]=='=' || tolower(read[i])==tolower(referencegenome[refpos])) {
						summaryTable[baseQuality][NONDBSNP_MATCH_COLUMN]++;  
					} else {
						summaryTable[baseQuality][NONDBSNP_MISMATCH_COLUMN]++; //
					}
					continue;   
                    
				}

				if (mask[refpos] == 2 ) {   
                                        
                    if (snpPos2Index[refpos] <= 0) { continue;}
                    
                    covarray[snpPos2Index[refpos]] ++; // calcuate the snp position covered by reads
                    
                                       									
                    allele = currentSample->markers[snpPos2Index[refpos]];    
                    
                    //printf("mask == 2\t%d\t%d\t%d\t%d\n",i,snpPos2Index[refpos], allele.isKnown(),allele.isHomozygous());                    
                    
					if(allele.isKnown() && allele.isHomozygous()) {
                        
                      //  printf("enter\n"); printf("%c\t%ld\n",referencegenome[refpos],refpos);
                        
						info = currentSample->GetMarkerInfo(snpPos2Index[refpos]);

                    // printf("enter1%c\n",allele[0]);
                        
					alleleNumber = info->GetAlleleLabel(allele[0])[0];
                    
                    //printf("%c\t%c\n",alleleNumber,referencegenome[refpos]);
                    
					readchar = tolower(read[i]); //printf("readchar= %c\n",readchar);
                    
                    char basechar = tolower(referencegenome[refpos]);
                    
                    if (nonrefhomo && (MatchAllele(basechar, alleleNumber)!=0))  continue;                                                    
                    //printf("here");                                                
                    
                    matchallele = MatchAllele(readchar, alleleNumber); 
                    
                    //printf("matchallele=%d\t%d\n",matchallele,baseQuality);
                    
                    if (verbose && matchallele != -1) {
                             
                    int chro = referencegenome.getChromosome(refpos);
                    int position = refpos - referencegenome.getChromosomeStart(chro)+1;
                    
                    fprintf (fout,"%s\t%d\t%d\t%d\t%c\t%c\t%c\n",(const char*) readID, chro+1,position,snpPos2Index[refpos]+1, toupper(readchar),referencegenome[refpos],toupper(alleleNumber));  
                    }                   
                                                   
                    if (matchallele)
                       { summaryTable[baseQuality][HAPMAP_MATCH_COLUMN]++; genocount ++;   maparray[snpPos2Index[refpos]][1] ++;}
	    			else if (matchallele==0) 
                       { summaryTable[baseQuality][HAPMAP_MISMATCH_COLUMN]++; genocount ++; maparray[snpPos2Index[refpos]][0] ++;}

                    
                  }
                  
                  
                }

			  
			} // end of for (int i = 0; i < read.Length(); i++)
             
            
        if (genocount >= maxcount)  { break;  }
                 
		} // end of while while (endFile)
		
        
        if (flagSAM) ifclose(f);
        else bamfile.Close();
        
              
        //if (count < READLOWERBOUND)  { printf ("LANE %s does not have enough reads!\n", (const char*) lanefile); continue; }
        
		
        for (int i = 0; i < 8; i ++) summaryCounts[i] = 0;
        
        PrintSummary(fout, summaryTable, summaryCounts, lanefile, fid, pid);
        
        if (coverage)
        {        
        int covcount = 0;
        for (int i = 0; i < ped.markerCount; i ++)
        { //fprintf(fout,"COVERAGE: %d\n",covarray[i]); 
          if (covarray[i]>0) covcount ++;}
        fprintf(fout,"COVERAGERESULT: %lf\n",1.0*covcount/ped.markerCount);
        }
		
		if (countbysite) 
        {    for (int i = 0; i < ped.markerCount; i ++)
            { if((maparray[i][0]+maparray[i][1])>0)  fprintf(fout,"countbysite:\t%s\t%d\t%d\t%d\t%d\t%d\t%lf\n",(const char*) lanefile,i,Pedigree::GetMarkerInfo(i)->chromosome,                    (int)round(Pedigree::GetMarkerInfo(i)->position * 100), maparray[i][0],maparray[i][1],1.0*maparray[i][0]/(maparray[i][0]+maparray[i][1])); }
        }     
        
	}  // end of while (!ifeof(fkey)) 
	
    
    delete [] summaryCounts;
    delete [] covarray;
    FreeDoubleMatrix(summaryTable, 300);
	mask.close();
	fclose(fout);
    ifclose(fkey);
    
}
