#ifndef _READBAM_H_
#define _READBAM_H_

/*
You could probably do a single pass through the pedigree to index all the individual ids 
(and ignore family ids). Thus:

StringIntHash pidIndex;

for (int i = 0; i < ped.count; i++)
   pidIndex.SetInteger(ped[i].pid, i);

After that, you can look up any individual using:

int i = pidIndex.Integer(pid);

You can also check for duplicated pids using:

if (ped.count > pidIndex.Entries())
   printf("Warning: Some individual ids are duplicated in your pedigree"
          "file. We recommend that every sample should have a unique id");

Goncalo

*/

/*
// /home/zhanxw/samtools-0.1.4/libsamtools.a
*/
        
#include <stdio.h>
#include <unistd.h>
#include "bam.h"
#include "sam.h"
#include "StringArray.h"


class CigarCol
{
    public:
    int *fieldCount;
    int size;
    String fieldTypes;
    String tmp;
    
    CigarCol() { fieldCount=NULL;};
    ~CigarCol(){  };
    
    void CigarParse(String cigar) {
     size = 0;   
        for  (int i = 0; i < cigar.Length(); i++)
        { if (!isdigit(cigar[i])) size ++;
        }
        
   
   fieldCount = new int[size];
    
  int k = 0; tmp = "";
  
  for  (int i = 0; i < cigar.Length(); i++)
  { if (isdigit(cigar[i])) tmp += cigar[i];
    else { fieldCount[k++] = tmp.AsInteger(); tmp.Clear(); fieldTypes += cigar[i];}
    }
  };
  
  void CigarClear(){
      
      if (fieldCount != NULL) { delete [] fieldCount; fieldTypes=""; tmp="";}
         
     };
       
    
};



class SAMEntry
{
public:
 String QNAME;
 unsigned int FLAG;
 String RNAME;
 unsigned int POS;
 unsigned int MAPQ;
 String CIGAR;
 String MRNM;
 unsigned int MPOS;
 unsigned int ISIZE;
 String SEQ;
 String QUAL;
 StringArray AUX;

 void Clear(void){
   QNAME="";
   FLAG=0;
   RNAME="";
   POS=0;
   MAPQ=0;
   CIGAR="";
   MRNM="";
   MPOS=0;
   ISIZE=0;
   SEQ="";
   QUAL="";
   AUX.Clear();
 };
};
class SAMFile
{
public:
 SAMFile(){ samheader= &samheader_s; bam1 = bam_init1();};
//   ~SAMFile();
 int Open(String& fileName)
 {
   samfile=samopen(fileName.c_str(), "rb", samheader);  //if .bam change "r" to "rb"
   if (samfile == 0)
   {
     fprintf(stderr, "Cannot open BAM file: %s !\n", fileName.c_str());
     return -1;
   }
   return 0;
 };
 int ReadLine(SAMEntry& samEntry)
 {
   bam1_t* b=bam1;
   int readByte=samread(samfile, b);
   if (readByte <= 0) return readByte;

   // format b and store values in s
   // modify code from bam.c : char *bam_format1(const bam_header_t *header, const bam1_t *b)
   bam_header_t* header = samfile->header;

   uint8_t *s = bam1_seq(b), *t = bam1_qual(b);
   int i;
   const bam1_core_t *c = &b->core;
//     kstring_t str;
//     str.l = str.m = 0; str.s = 0;

   //ksprintf(&str, "%s\t%d\t", bam1_qname(b), c->flag);
   samEntry.QNAME=(bam1_qname(b));
   samEntry.FLAG = c->flag;

   if (c->tid < 0)
     samEntry.RNAME="*"; //kputs("*\t", &str);
   else //ksprintf(&str, "%s\t", header->target_name[c->tid]);
     samEntry.RNAME.printf("%s", header->target_name[c->tid]);
//     ksprintf(&str, "%d\t%d\t", c->pos + 1, c->qual);
   samEntry.POS = c->pos + 1;
   samEntry.MAPQ = c->qual;
   if (c->n_cigar == 0) samEntry.CIGAR="*"; //kputc('*', &str);
   else {
     String tag;
     for (i = 0; i < c->n_cigar; ++i)
     {
       //      ksprintf(&str, "%d%c",bam1_cigar(b)[i]>>BAM_CIGAR_SHIFT, "MIDNSHP"[bam1_cigar(b)[i]&BAM_CIGAR_MASK]);
       tag.printf("%d%c",bam1_cigar(b)[i]>>BAM_CIGAR_SHIFT,"MIDNSHP"[bam1_cigar(b)[i]&BAM_CIGAR_MASK]);
       samEntry.CIGAR+=tag;
     }
   }
//     kputc('\t', &str);

   if (c->mtid < 0) samEntry.MRNM="*"; //kputs("*\t", &str);
   else if (c->mtid == c->tid) samEntry.MRNM="="; //kputs("=\t", &str);
   else samEntry.MRNM.printf("%s", header->target_name[c->mtid]);
//ksprintf(&str, "%s\t", header->target_name[c->mtid]);

//     ksprintf(&str, "%d\t%d\t", c->mpos + 1, c->isize);
   samEntry.MPOS=c->mpos + 1;
   samEntry.ISIZE=c->isize;

   samEntry.SEQ.Clear();
   for (i = 0; i < c->l_qseq; ++i)
samEntry.SEQ.Add(bam_nt16_rev_table[bam1_seqi(s, i)]);
//kputc(bam_nt16_rev_table[bam1_seqi(s, i)], &str);

//     kputc('\t', &str);

   samEntry.QUAL.Clear();
   if (t[0] == 0xff) samEntry.QUAL="*"; //kputc('*', &str);
   else for (i = 0; i < c->l_qseq; ++i) samEntry.QUAL.Add(t[i] + 33);
 // kputc(t[i] + 33, &str);


   samEntry.AUX.Clear();
   s = bam1_aux(b);
   while (s < b->data + b->data_len) {
     String tag;
     uint8_t type, key[2];
     key[0] = s[0]; key[1] = s[1];
     s += 2; type = *s; ++s;
     // ksprintf(&str, "\t%c%c:", key[0], key[1]);
     // tag.printf("%c%c:", key[0], key[1]);

     if (type == 'A')      { tag.printf("%c%c:A:%c", key[0], key[1],*s); ++s; }
     else if (type == 'C') { tag.printf("%c%c:i:%u", key[0], key[1],*s); ++s; }
     else if (type == 'c') { tag.printf("%c%c:i:%d", key[0], key[1],*s); ++s; }
     else if (type == 'S') { tag.printf("%c%c:i:%u", key[0], key[1],*(uint16_t*)s);s += 2; }
     else if (type == 's') { tag.printf("%c%c:i:%d", key[0], key[1],*(int16_t*)s);  s += 2; }
     else if (type == 'I') { tag.printf("%c%c:i:%u", key[0], key[1],*(uint32_t*)s); s += 4; }
     else if (type == 'i') { tag.printf("%c%c:i:%d", key[0], key[1],*(int32_t*)s); s += 4; }
     else if (type == 'f') { tag.printf("%c%c:f:%g", key[0], key[1],*(float*)s); s += 4; }
     else if (type == 'd') { tag.printf("%c%c:d:%lg", key[0],key[1],*(double*)s); s += 8;}
     else if (type == 'Z' || type == 'H') {
       tag.printf("%c%c:%c:", key[0], key[1],type);
       while (*s) tag.Add(*s++);  ++s; }
     samEntry.AUX.Push(tag);
   }
//     return str.s;

   return readByte;
 };
 void Close()
 {
   if (samfile != 0)
     samclose(samfile);
 };

private:
 samfile_t* samfile;
 bam_header_t samheader_s;
 bam_header_t* samheader;
 bam1_t* bam1;
};

#endif

